"""
- `UNAUTHORISED_REQUEST_ERROR_MESSAGE` (str): Plain text error message.
  Defaults to "Unauthorised request. {0}".
- `UNAUTHORISED_REQUEST_ERROR_TEMPLATE` (str): Path to 401 template that
  should be rendered in case of 401 responses. Defaults to empty string (not
  provided).
- `AUTH_USER` (str): Default ``auth_user`` for ``ska.sign_url`` function.
  Defaults to "ska-auth-user".
- `USER_GET_CALLBACK` (str): User get callback (when user is fetched in auth
  backend).
- `USER_VALIDATE_CALLBACK` (str): User validate callback (fired before user is
  created; created to allow custom logic to the user authentication before
  user object is even created).
- `USER_CREATE_CALLBACK` (str): User create callback (when user is created in
  auth backend).
- `USER_INFO_CALLBACK` (str): User info callback.
- `REDIRECT_AFTER_LOGIN` (str): Redirect after login.
- `DB_STORE_SIGNATURES` (bool): If set to True, signatures are stored in the
  database.
- `DB_PERFORM_SIGNATURE_CHECK` (bool): If set to True, an extra check is
  fired on whether the token has already been used or not.
- `PROVIDERS` (dict): A dictionary where key is the provider UID and the key
  is another dictionary holding the following provider specific keys:
  'SECRET_KEY', 'USER_GET_CALLBACK', 'USER_CREATE_CALLBACK',
  'USER_INFO_CALLBACK', 'REDIRECT_AFTER_LOGIN'. Note, that the 'SECRET_KEY'
  is a required key. The rest are optional, and if given, override
  respectively the values of ``ska.contrib.django.ska.settings``.
"""

from __future__ import absolute_import

from ska.gettext import _

__title__ = 'ska.contrib.django.ska.defaults'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2013-2018 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'AUTH_USER',
    'DB_PERFORM_SIGNATURE_CHECK',
    'DB_STORE_SIGNATURES',
    'PROVIDERS',
    'REDIRECT_AFTER_LOGIN',
    'UNAUTHORISED_REQUEST_ERROR_MESSAGE',
    'UNAUTHORISED_REQUEST_ERROR_TEMPLATE',
    'USER_CREATE_CALLBACK',
    'USER_GET_CALLBACK',
    'USER_INFO_CALLBACK',
    'USER_VALIDATE_CALLBACK',
)

AUTH_USER = 'ska-auth-user'

UNAUTHORISED_REQUEST_ERROR_MESSAGE = _("Unauthorised request. {0}")

UNAUTHORISED_REQUEST_ERROR_TEMPLATE = ''

USER_VALIDATE_CALLBACK = None
USER_GET_CALLBACK = None
USER_CREATE_CALLBACK = None
USER_INFO_CALLBACK = None
REDIRECT_AFTER_LOGIN = ''

DB_STORE_SIGNATURES = False
DB_PERFORM_SIGNATURE_CHECK = False

PROVIDERS = {}



"""
- ``validate_signed_request``: Function decorator. Validate request signature.
  Applies appropriate validation mechanism to the request data. Assumes
  ``SKA_SECRET_KEY`` to be in ``settings`` module.

  Arguments to be used with `ska.validate_signed_request_data` shortcut
  function.

  :param str secret_key: The shared secret key.
  :param str signature_param: Name of the (for example GET or POST) param name
      which holds the ``signature`` value.
  :param str auth_user_param: Name of the (for example GET or POST) param name
      which holds the ``auth_user`` value.
  :param str valid_until_param: Name of the (foe example GET or POST) param
      name which holds the ``valid_until`` value.

- ``sign_url``: Method decorator (to be used in models). Signs the URL.

  Arguments to be used with `ska.sign_url` shortcut function.

  :param str auth_user: Username of the user making the request.
  :param str secret_key: The shared secret key.
  :param float|str valid_until: Unix timestamp. If not given, generated
      automatically (now + lifetime).
  :param int lifetime: Signature lifetime in seconds.
  :param str suffix: Suffix to add after the ``endpoint_url`` and before the
      appended signature params.
  :param str signature_param: Name of the GET param name which would hold the
      generated signature value.
  :param str auth_user_param: Name of the GET param name which would hold the
      ``auth_user`` value.
  :param str valid_until_param: Name of the GET param name which would hold
      the ``valid_until`` value.
"""

from __future__ import absolute_import

from django.shortcuts import render
from django.utils.translation import ugettext, ugettext_lazy as _

from nine import versions

from six import PY3, text_type

from .... import validate_signed_request_data, sign_url as ska_sign_url
from ....defaults import (
    SIGNATURE_LIFETIME,
    DEFAULT_URL_SUFFIX,
    DEFAULT_SIGNATURE_PARAM,
    DEFAULT_AUTH_USER_PARAM,
    DEFAULT_VALID_UNTIL_PARAM,
    DEFAULT_EXTRA_PARAM,
)

from .http import HttpResponseUnauthorized
from .settings import (
    SECRET_KEY,
    AUTH_USER,
    UNAUTHORISED_REQUEST_ERROR_MESSAGE,
    UNAUTHORISED_REQUEST_ERROR_TEMPLATE,
)

__title__ = 'ska.contrib.django.ska.decorators'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2013-2018 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'validate_signed_request',
    'sign_url',
    'BaseValidateSignedRequest',
    'ValidateSignedRequest',
    'MethodValidateSignedRequest',
    'SignAbsoluteURL',
)


class BaseValidateSignedRequest(object):
    """BaseValidateSignedRequest."""

    def __init__(self, secret_key=SECRET_KEY,
                 signature_param=DEFAULT_SIGNATURE_PARAM,
                 auth_user_param=DEFAULT_AUTH_USER_PARAM,
                 valid_until_param=DEFAULT_VALID_UNTIL_PARAM,
                 extra_param=DEFAULT_EXTRA_PARAM):
        """Constructor."""
        self.secret_key = secret_key
        self.signature_param = signature_param
        self.auth_user_param = auth_user_param
        self.valid_until_param = valid_until_param
        self.extra_param = extra_param


class ValidateSignedRequest(BaseValidateSignedRequest):
    """ValidateSignedRequest.

    Function decorator. Validate request signature. Applies appropriate
    validation mechanism to the request data. Assumes ``SKA_SECRET_KEY`` to be
    in ``settings`` module.

    Arguments to be used with `ska.validate_signed_request_data` shortcut
    function.

    :attribute str secret_key: The shared secret key.
    :attribute str signature_param: Name of the (for example GET or POST) param
        name which holds the ``signature`` value.
    :attribute str auth_user_param: Name of the (for example GET or POST) param
        name which holds the ``auth_user`` value.
    :attribute str valid_until_param: Name of the (foe example GET or POST)
        param name which holds the ``valid_until`` value.
    :attribute str extra_param: Name of the (foe example GET or POST) param
        name which holds the ``extra`` value.

    :example:

    >>> from ska.contrib.django.ska.decorators import validate_signed_request
    >>>
    >>> @validate_signed_request()
    >>> def detail(request, slug, template_name='foo/detail.html'):
    >>>     # Your code
    """

    def __call__(self, func):
        """Call."""
        def inner(request, *args, **kwargs):
            """Inner."""
            # Validating the request.
            if versions.DJANGO_GTE_1_7:
                request_data = request.GET.dict()
            else:
                request_data = request.REQUEST
            validation_result = validate_signed_request_data(
                data=request_data,
                secret_key=self.secret_key,
                signature_param=self.signature_param,
                auth_user_param=self.auth_user_param,
                valid_until_param=self.valid_until_param,
                extra_param=self.extra_param
            )
            if validation_result.result is True:
                # If validated, just return the func as is.
                return func(request, *args, **kwargs)
            else:
                # Otherwise...
                if UNAUTHORISED_REQUEST_ERROR_TEMPLATE:
                    # If template to display the error message is set in
                    # ska (django-ska) settings, use it to render the message
                    # and return ``HttpResponseUnauthorized`` response
                    # describing the error.
                    response_content = render(
                        request,
                        UNAUTHORISED_REQUEST_ERROR_TEMPLATE,
                        {'reason': '; '.join(validation_result.reason)}
                    )
                    return HttpResponseUnauthorized(response_content)
                else:
                    # Otherwise, return plain text message with describing the
                    # error.
                    return HttpResponseUnauthorized(
                        ugettext(UNAUTHORISED_REQUEST_ERROR_MESSAGE).format(
                            '; '.join(validation_result.reason)
                        )
                    )
        return inner


validate_signed_request = ValidateSignedRequest


class MethodValidateSignedRequest(BaseValidateSignedRequest):
    """MethodValidateSignedRequest.

    Method decorator. Validate request signature. Applies appropriate
    validation mechanism to the request data. Assumes ``SKA_SECRET_KEY`` to be
    in ``settings`` module.

    Arguments to be used with `ska.validate_signed_request_data` shortcut
    function.

    :attribute str secret_key: The shared secret key.
    :attribute str signature_param: Name of the (for example GET or POST) param
        name which holds the ``signature`` value.
    :attribute str auth_user_param: Name of the (for example GET or POST) param
        name which holds the ``auth_user`` value.
    :attribute str valid_until_param: Name of the (foe example GET or POST)
        param name which holds the ``valid_until`` value.
    :attribute str extra_param: Name of the (foe example GET or POST) param
        name which holds the ``extra`` value.

    :example:

    >>> from ska.contrib.django.ska.decorators import m_validate_signed_request
    >>>
    >>> class FooDetailView(View):
    >>>     @validate_signed_request()
    >>>     def get(self, request, slug, template_name='foo/detail.html'):
    >>>         # Your code
    """

    def __call__(self, func):
        """Call."""
        def inner(this, request, *args, **kwargs):
            """Inner."""
            if versions.DJANGO_GTE_1_7:
                request_data = request.GET.dict()
            else:
                request_data = request.REQUEST
            # Validating the request.
            validation_result = validate_signed_request_data(
                data=request_data,
                secret_key=self.secret_key,
                signature_param=self.signature_param,
                auth_user_param=self.auth_user_param,
                valid_until_param=self.valid_until_param,
                extra_param=self.extra_param
            )
            if validation_result.result is True:
                # If validated, just return the func as is.
                return func(this, request, *args, **kwargs)
            else:
                # Otherwise...
                if UNAUTHORISED_REQUEST_ERROR_TEMPLATE:
                    # If template to display the error message is set in
                    # ska (django-ska) settings, use it to render the message
                    # and return ``HttpResponseUnauthorized`` response
                    # describing the error.
                    response_content = render(
                        request,
                        UNAUTHORISED_REQUEST_ERROR_TEMPLATE,
                        {'reason': '; '.join(validation_result.reason)}
                    )
                    return HttpResponseUnauthorized(response_content)
                else:
                    # Otherwise, return plain text message with describing the
                    # error.
                    return HttpResponseUnauthorized(
                        ugettext(UNAUTHORISED_REQUEST_ERROR_MESSAGE).format(
                            '; '.join(validation_result.reason)
                        )
                    )
        return inner


m_validate_signed_request = MethodValidateSignedRequest


class SignAbsoluteURL(object):
    """SignAbsoluteURL.

    Method decorator (to be used in models). Signs the URL.

    Arguments to be used with `ska.sign_url` shortcut function.

    :attribute str auth_user: Username of the user making the request.
    :attribute str secret_key: The shared secret key.
    :attribute float|str valid_until: Unix timestamp. If not given, generated
        automatically (now + lifetime).
    :attribute int lifetime: Signature lifetime in seconds.
    :attribute str suffix: Suffix to add after the ``endpoint_url`` and before
        the appended signature params.
    :attribute str signature_param: Name of the GET param name which would hold
        the generated signature value.
    :attribute str auth_user_param: Name of the GET param name which would hold
        the ``auth_user`` value.
    :attribute str valid_until_param: Name of the GET param name which would
        hold the ``valid_until`` value.
    :attribute dict extra: Dict of extra params to append to signed URL.
    :attribute str extra_param: Name of the GET param name which would hold
        the ``extra`` value.

    :example:

    >>> from ska.contrib.django.ska.decorators import sign_url
    >>>
    >>> class FooItem(models.Model):
    >>>     title = models.CharField(_("Title"), max_length=100)
    >>>     slug = models.SlugField(unique=True, verbose_name=_("Slug"))
    >>>     body = models.TextField(_("Body"))
    >>>
    >>>     @sign_url()
    >>>     def get_signed_absolute_url(self):
    >>>         return reverse('foo.detail', kwargs={'slug': self.slug})
    """

    def __init__(self, auth_user=AUTH_USER, secret_key=SECRET_KEY,
                 valid_until=None, lifetime=SIGNATURE_LIFETIME,
                 suffix=DEFAULT_URL_SUFFIX,
                 signature_param=DEFAULT_SIGNATURE_PARAM,
                 auth_user_param=DEFAULT_AUTH_USER_PARAM,
                 valid_until_param=DEFAULT_VALID_UNTIL_PARAM,
                 extra=None, extra_param=DEFAULT_EXTRA_PARAM):
        """Constructor."""
        self.auth_user = auth_user
        self.secret_key = secret_key
        self.valid_until = valid_until
        self.lifetime = lifetime
        self.suffix = suffix
        self.signature_param = signature_param
        self.auth_user_param = auth_user_param
        self.valid_until_param = valid_until_param
        self.extra = extra if extra is not None else {}
        self.extra_param = extra_param

    def __call__(self, func):
        """Call."""
        def inner(this, *args, **kwargs):
            """Inner."""
            if not PY3:
                url = text_type(func(this, *args, **kwargs))
            else:
                url = func(this, *args, **kwargs)

            return ska_sign_url(
                auth_user=self.auth_user,
                secret_key=self.secret_key,
                valid_until=self.valid_until,
                lifetime=self.lifetime,
                url=url,
                suffix=self.suffix,
                signature_param=self.signature_param,
                auth_user_param=self.auth_user_param,
                valid_until_param=self.valid_until_param,
                extra=self.extra,
                extra_param=self.extra_param
            )
        return inner


sign_url = SignAbsoluteURL



from .constance_views import *
from .default_views import *

__title__ = 'ska.contrib.django.ska.views'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2013-2018 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'login',
    'constance_login',
)



from __future__ import absolute_import

__title__ = 'ska.contrib.django.ska.apps'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2013-2018 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = ('default_app_config',)

default_app_config = 'ska.contrib.django.ska.apps.Config'



__title__ = 'ska.gettext'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2013-2018 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = ('_',)


def _(val):
    """For future integrations with gettext."""
    return val



from six import python_2_unicode_compatible

from .gettext import _

__title__ = 'ska.error_codes'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2013-2018 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'ErrorCode',
    'INVALID_SIGNATURE',
    'SIGNATURE_TIMESTAMP_EXPIRED',
)


@python_2_unicode_compatible
class ErrorCode(object):
    """Base error code.

    If you have ever used the following code with `validation_result`:

    >>> human_readable_error = ' '.join(validation_result.reason)

    ...change it as follows:

    >>> human_readable_error = validation_result.message

    :property int code: Just an integer code.
    :property string message: Human readable represantation of the error
        message.
    """
    __slots__ = ('code', 'message')

    def __init__(self, code, message):
        self.code = code
        self.message = message

    def __str__(self):
        return self.message
    __repr__ = __str__

    def __int__(self):
        return self.code


INVALID_SIGNATURE = ErrorCode(1, _("Invalid signature!"))
SIGNATURE_TIMESTAMP_EXPIRED = ErrorCode(2, _("Signature timestamp expired!"))



from __future__ import absolute_import, print_function

import datetime
import logging
import os
import time
import unittest

import pytest

from .helpers import (
    project_dir,
    PROJECT_DIR,
    LOG_INFO,
    log_info,
    change_date,
    NUM_ITEMS,
)

if os.environ.get("DJANGO_SETTINGS_MODULE", None):
    from .helpers import (
        create_admin_user,
        generate_data,
    )

__title__ = 'ska.contrib.django.ska.tests.test_django_ska'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2013-2018 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'

logger = logging.getLogger(__name__)


# *********************************************************************
# *********************************************************************
# *********************************************************************

# Skipping from non-Django tests.
if os.environ.get("DJANGO_SETTINGS_MODULE", None):

    from django.core import mail
    from django.core.management import call_command
    from django.test import Client, TransactionTestCase

    from ska import sign_url
    from ska.defaults import DEFAULT_PROVIDER_PARAM
    from ska.contrib.django.ska.settings import (
        SECRET_KEY,
        DB_STORE_SIGNATURES,
        DB_PERFORM_SIGNATURE_CHECK,
        PROVIDERS
    )
    from ska.contrib.django.ska.models import Signature

    import factories

    SKA_TEST_USER_USERNAME = factories.TEST_ADMIN_USERNAME
    SKA_TEST_USER_PASSWORD = factories.TEST_PASSWORD


    @pytest.mark.django_db
    class SkaDecoratorsTest(TransactionTestCase):
        """Testing model- and view- decorators."""

        pytestmark = pytest.mark.django_db

        def setUp(self):
            items = generate_data(NUM_ITEMS)

            # Testing the URLs
            self.item = items[0]

        @log_info
        def test_01_model_decorator(self):
            """Test the ``ska.contrib.django.ska.decorators.sign_url``."""
            # Testing signed URLs
            signed_absolute_url = self.item.get_signed_absolute_url()
            self.assertIsNotNone(signed_absolute_url)
            return signed_absolute_url

        @log_info
        def test_02_view_decorator_with_signed_url(self):
            """Test view decorator with signed URL.

            Test ``ska.contrib.django.ska.decorators.validate_signed_request``.
            """
            flow = []

            # Testing signed URLs
            signed_absolute_url = self.item.get_signed_absolute_url()
            self.assertIsNotNone(signed_absolute_url)
            flow.append(('Signed absolute URL', signed_absolute_url))

            # Testing view with signed URL
            client = Client()
            response = client.get(signed_absolute_url, {})
            response_status_code = getattr(response, 'status_code', None)
            self.assertIn(response_status_code, (200, 201, 202))
            flow.append(
                ('Response status code for signed URL', response_status_code)
            )

            return flow

        @log_info
        def test_03_view_decorator_with_unsigned_url(self):
            """Test view decorator with unsigned URL.

            Test the
            ``ska.contrib.django.ska.decorators.validate_signed_request`` view
            decorator with unsigned URL.
            """
            flow = []

            # Testing unsigned URLs
            absolute_url = self.item.get_absolute_url()
            self.assertTrue(absolute_url is not None)
            flow.append(('Unsigned absolute URL', absolute_url))

            # Testing view with signed URL
            client = Client()
            response = client.get(absolute_url, {})
            response_status_code = getattr(response, 'status_code', None)
            response_content = getattr(response, 'content', "")
            self.assertIn(response_status_code, (401,))
            flow.append(
                ('Response status code for unsigned URL', response_status_code)
            )
            flow.append(
                ('Response content for unsigned URL', response_content)
            )

            return flow

        @log_info
        def test_04_class_based_view_decorator_with_signed_url(self):
            """Test class based view decorator with signed URL.

            Test ``ska.contrib.django.ska.decorators.validate_signed_request``.
            """
            flow = []

            # Testing signed URLs
            signed_absolute_url = self.item \
                .get_signed_class_based_absolute_url()
            self.assertIsNotNone(signed_absolute_url)
            flow.append(('Signed absolute URL', signed_absolute_url))

            # Testing view with signed URL
            client = Client()
            response = client.get(signed_absolute_url, {})
            response_status_code = getattr(response, 'status_code', None)
            self.assertIn(response_status_code, (200, 201, 202))
            flow.append(
                ('Response status code for signed URL', response_status_code)
            )

            return flow

        @log_info
        def test_05_class_based_view_decorator_with_unsigned_url(self):
            """Test class based view decorator with unsigned URL.

            Test the
            ``ska.contrib.django.ska.decorators.validate_signed_request`` view
            decorator with unsigned URL.
            """
            flow = []

            # Testing unsigned URLs
            absolute_url = self.item.get_cbv_absolute_url()
            self.assertTrue(absolute_url is not None)
            flow.append(('Unsigned absolute URL', absolute_url))

            # Testing view with signed URL
            client = Client()
            response = client.get(absolute_url, {})
            response_status_code = getattr(response, 'status_code', None)
            response_content = getattr(response, 'content', "")
            self.assertIn(response_status_code, (401,))
            flow.append(
                ('Response status code for unsigned URL', response_status_code)
            )
            flow.append(
                ('Response content for unsigned URL', response_content)
            )

            return flow


    @pytest.mark.django_db
    class SkaAuthenticationBackendTest(TransactionTestCase):
        """Tests for auth backend."""

        pytestmark = pytest.mark.django_db

        def setUp(self):
            self.AUTH_USER = 'test_auth_backend_user'
            self.AUTH_USER_EMAIL = 'test_ska_auth_user@mail.example.com'
            self.AUTH_USER_FIRST_NAME = 'John'
            self.AUTH_USER_LAST_NAME = 'Doe'
            self.PROVIDER_NAME = 'client_1.admins'
            self.LOGIN_URL = '/ska/login/'

        def __test_login(self,
                         secret_key,
                         response_codes,
                         provider_name=None,
                         logout=False,
                         test_email_data=None,
                         do_signature_check=True,
                         auth_user=None,
                         auth_user_email=None,
                         debug_info=""):
            """Test login.

            :param secret_key:
            :param response_codes: First response code is used
            :param provider_name:
            :param logout:
            :param test_email_data:
            :param do_signature_check: It's a tuple. The second code is
                used only if both `DB_STORE_SIGNATURES` and
                `DB_PERFORM_SIGNATURE_CHECK` values are True, which means
                that users can't use same signatures twice.
            :param auth_user:
            :param auth_user_email:
            :return:
            """
            flow = []

            if not auth_user:
                auth_user = self.AUTH_USER
            if not auth_user_email:
                auth_user_email = self.AUTH_USER_EMAIL

            first_response_code, second_response_code = response_codes

            # Testing signed URLs
            extra = {
                'email': auth_user_email,
                'first_name': self.AUTH_USER_FIRST_NAME,
                'last_name': self.AUTH_USER_LAST_NAME,
            }
            if provider_name:
                extra.update({DEFAULT_PROVIDER_PARAM: provider_name})

            signed_login_url = sign_url(
                auth_user=auth_user,
                secret_key=secret_key,
                url=self.LOGIN_URL,
                extra=extra
            )

            self.assertIsNotNone(signed_login_url)
            flow.append(('Signed login URL', signed_login_url))

            # Testing view with signed URL
            self._client = Client()
            response = self._client.get(signed_login_url, {})
            response_status_code = getattr(response, 'status_code', None)

            self.assertIn(response_status_code, (first_response_code,))
            flow.append(
                (
                    'Response status code for signed URL',
                    response_status_code
                )
            )

            if test_email_data:
                self.assertEqual(
                    len(mail.outbox),
                    len(test_email_data['emails'])
                )

                for _index, _subject in test_email_data['emails'].items():
                    self.assertEqual(
                        mail.outbox[_index].subject,
                        _subject
                    )

            if logout:
                self.__logout()

            # If both `DB_STORE_SIGNATURES` and `DB_PERFORM_SIGNATURE_CHECK`
            # are set to True, second login attempt shall not be successful.
            if DB_STORE_SIGNATURES \
                    and DB_PERFORM_SIGNATURE_CHECK \
                    and do_signature_check:
                # Testing again with signed URL and this time, it should fail

                self._client = Client()
                response = self._client.get(signed_login_url, {})
                response_status_code = getattr(response, 'status_code', None)
                self.assertIn(response_status_code, (second_response_code,))
                flow.append(
                    (
                        'Response status '
                        'code for signed URL', response_status_code
                    )
                )
                if logout:
                    self.__logout()

            return flow

        def __logout(self):
            """Log out."""
            self._client.logout()

        @log_info
        def test_01_login(self):
            """Test auth using general ``SECRET_KEY``."""
            return self.__test_login(
                SECRET_KEY,
                [302, 403],
                logout=True,
                debug_info="test_01_login"
            )

        @log_info
        def test_02_provider_login(self):
            """Test auth using ``SECRET_KEY`` defined in ``PROVIDERS``."""
            secret_key = PROVIDERS[self.PROVIDER_NAME]['SECRET_KEY']

            # Authenticate for the first time. There shall be 2 emails
            # for `create` and `info` callbacks.
            self.__test_login(
                secret_key,
                [302, 403],
                self.PROVIDER_NAME,
                logout=True,
                test_email_data={
                    'emails': {
                        0: 'Welcome validate::admins!',
                        1: 'Welcome create::admins!',
                        2: 'Welcome info::admins!',
                    }
                },
                debug_info="test_02_provider_login::first time"
            )

            # Sleep for just one second between first and the second tests.
            time.sleep(1)

            # Authenticate for the second time. There shall be still 2 emails
            # from the first time login (since we're in the same test method
            # still) and 2 new ones for `get` and `info` callbacks.
            self.__test_login(
                secret_key,
                [302, 403],
                self.PROVIDER_NAME,
                logout=True,
                test_email_data={
                    'emails': {
                        # First round
                        0: 'Welcome validate::admins!',
                        1: 'Welcome create::admins!',
                        2: 'Welcome info::admins!',
                        # Second round
                        # TODO: this is weird. validate::admin shall not
                        # appear here twice.
                        3: 'Welcome validate::admins!',
                        4: 'Welcome validate::admins!',
                        5: 'Welcome get::admins!',
                        6: 'Welcome info::admins!',
                    }
                },
                do_signature_check=False,
                debug_info="test_02_provider_login::second time"
            )

        @log_info
        def test_03_login_fail_wrong_secret_key(self):
            """Fail test auth using general ``SECRET_KEY``.

            Fail test auth using general ``SECRET_KEY`` providing wrong
            secret key.
            """
            return self.__test_login(
                SECRET_KEY + 'wrong',
                [403, 403],
                logout=True,
                debug_info="test_03_login_fail_wrong_secret_key"
            )

        @log_info
        def test_04_provider_login_fail_wrong_secret_key(self):
            """Fail test authentication.

            Fail test authentication using ``SECRET_KEY`` defined in `
            `PROVIDERS``, providing wrong secret key.
            """
            secret_key = PROVIDERS[self.PROVIDER_NAME]['SECRET_KEY']
            return self.__test_login(
                secret_key + 'wrong',
                [403, 403],
                self.PROVIDER_NAME,
                logout=True
            )

        @log_info
        def test_05_provider_login_fail_wrong_provider(self):
            """Test provider login fail wrong provider.

            Fail test authentication using ``SECRET_KEY`` defined in
            ``PROVIDERS``, providing wrong provider name.
            """
            secret_key = PROVIDERS[self.PROVIDER_NAME]['SECRET_KEY']
            return self.__test_login(
                secret_key + 'wrong',
                [403, 403],
                self.PROVIDER_NAME + 'wrong',
                logout=True,
                debug_info="test_05_provider_login_fail_wrong_provider"
            )

        @log_info
        def test_06_purge_stored_signatures_data(self):
            """Test purge stored signature data."""
            secret_key = PROVIDERS[self.PROVIDER_NAME]['SECRET_KEY']
            # Login
            self.__test_login(
                secret_key,
                [302, 403],
                self.PROVIDER_NAME,
                logout=True,
                debug_info="test_06_purge_stored_signatures_data"
            )

            # Duplicate signatures
            signature = Signature.objects.first()
            signature.id = None
            signature.signature += '0'
            signature.save()
            signature.id = None
            signature.signature += '0'
            signature.save()
            # There should be 3
            self.assertEqual(Signature.objects.all().count(), 3)

            # Manually set valid_until to no longer valid so that we can
            # test
            invalid_until = (
                datetime.datetime.now() - datetime.timedelta(minutes=20)
            )
            Signature.objects.update(
                valid_until=invalid_until
            )

            # Call purge command
            call_command('ska_purge_stored_signature_data')
            # All old records shall be gone
            self.assertEqual(Signature.objects.all().count(), 0)

        @log_info
        def test_07_provider_login_forbidden_email(self):
            """Test auth using ``SECRET_KEY`` defined in ``PROVIDERS``."""
            secret_key = PROVIDERS[self.PROVIDER_NAME]['SECRET_KEY']

            # Authenticate for the first time. There shall be 2 emails
            # for `create` and `info` callbacks.
            self.__test_login(
                secret_key,
                [403, 403],
                self.PROVIDER_NAME,
                logout=True,
                test_email_data={
                    'emails': {}
                },
                auth_user_email='forbidden@example.com',
                debug_info="test_07_provider_login_forbidden_email"
            )

        @log_info
        def test_08_provider_login_forbidden_username(self):
            """Test auth using ``SECRET_KEY`` defined in ``PROVIDERS``."""
            secret_key = PROVIDERS[self.PROVIDER_NAME]['SECRET_KEY']

            # Authenticate for the first time. There shall be 2 emails
            # for `create` and `info` callbacks.
            self.__test_login(
                secret_key,
                [403, 403],
                self.PROVIDER_NAME,
                logout=True,
                test_email_data={
                    'emails': {}
                },
                auth_user='forbidden_username',
                debug_info="test_08_provider_login_forbidden_username"
            )


if __name__ == "__main__":
    # Tests
    unittest.main()
